"""Utility methods"""
from __future__ import annotations

import datetime as dt
from packaging import version
from .model import RointeProduct

DEFAULT_TIME_ZONE: dt.tzinfo = dt.timezone.utc


def now(time_zone=None) -> dt.datetime:
    """Get now in specified time zone."""
    return dt.datetime.now(time_zone or DEFAULT_TIME_ZONE)


def find_max_fw_version(data, device_class: str, product_version: str) -> str | None:
    """Finds the latest FW version for a specific device class and version"""

    if device_class in data:
        if (
            product_version in data[device_class]
            and "end_user" in data[device_class][product_version]
        ):
            root = data[device_class][product_version]["end_user"]

            max_version = None

            for entry in root:
                ptr = version.parse(entry)
                if max_version is None or ptr > max_version:
                    max_version = ptr

            return str(max_version)

    return None


def build_update_map(firmware_data: dict) -> dict[RointeProduct, dict[str, str]]:
    """
    Builds an update map for each device.

    Output is a dict of [product, [existing_version, target_version]

    Where [target_version] is the next version the product can be updated to.
    and [existing_verion] is the product's current version.
    """
    fw_map = {}

    for entry in RointeProduct:
        fw_map[entry] = build_product_fw_map(entry, firmware_data)

    return fw_map


def build_product_fw_map(product: RointeProduct, firmware_data: dict) -> dict[str, str]:
    """Builds the upgrade map for a specific product."""

    if product.device_type not in firmware_data:
        return None

    root_node = firmware_data[product.device_type]

    if product.version not in root_node:
        return None

    product_versions = root_node[product.version]["end_user"]

    upgrade_map = {}

    for version_entry in product_versions:
        new_version = product_versions[version_entry].get("firmware_new_version", None)

        if new_version:
            upgrade_map[version_entry] = new_version

    return upgrade_map


def get_product_by_type_version(
    product_type: str, product_version: str
) -> RointeProduct | None:
    """Find the product model by its type and version."""

    for entry in RointeProduct:
        if entry.device_type == product_type and entry.version == product_version:
            return entry

    return None



"""Data Model enumerations."""

from __future__ import annotations
from enum import Enum


class RointeProduct(Enum):
    """Device Models and Versions enum"""

    def __init__(self, product_name: str, device_type: str, version: str):
        """Initializes the enum."""
        self.product_name = product_name
        self.device_type = device_type
        self.version = version

    RADIATOR_V1 = "DeltaUltimate Radiator", "radiator", "v1"
    RADIATOR_V2 = "D-Series Radiator", "radiator", "v2"
    RADIATOR_B_V2 = "Belize/Olympia Radiator", "radiatorb", "v2"
    TOWEL_RAIL_V1 = "Towel v1", "towel", "v1"
    TOWEL_RAIL_V2 = "Towel Rail", "towel", "v2"
    WATER_HEATER_V1 = "Water Heater v1", "acs", "v1"
    WATER_HEATER_V2 = "Water Heater v2", "acs", "v2"
    THERMO_V2 = "Thermostat", "therm", "v2"
    OVAL_TOWEL = "Oval Towel", "oval_towel", "v2"



class DeviceMode(Enum):
    """Device working modes."""

    AUTO = "auto"
    MAN = "manual"


class ScheduleMode(Enum):
    """Radiator schedule modes."""

    COMFORT = "C"
    ECO = "E"
    NONE = "O"



"""Data objects"""

from __future__ import annotations
from datetime import datetime


class EnergyConsumptionData:
    """Data object for energy statistics."""

    start: datetime
    end: datetime
    kwh: float
    effective_power: float

    created: datetime

    def __init__(
        self,
        start: datetime,
        end: datetime,
        kwh: float,
        effective_power: float,
        created: datetime,
    ):
        """Initialize the value."""
        self.start = start
        self.end = end
        self.kwh = kwh
        self.effective_power = effective_power
        self.created = created



"""Device data model."""

from __future__ import annotations
from datetime import datetime

from rointesdk.utils import get_product_by_type_version

from .model import ScheduleMode, RointeProduct
from .dto import EnergyConsumptionData
from . import utils


class RointeDevice:
    """Represent a Rointe device from the API."""

    id: str
    name: str
    serialnumber: str
    type: str

    # This represents the model of a particular product. Not the FW version.
    product_version: str

    firmware_version: str
    latest_firmware_version: str

    nominal_power: int
    power: bool

    # Describes the preset: ICE, ECO, Comfort, None
    preset: str
    mode: str

    temp: float
    temp_calc: float
    temp_probe: float

    # preset temperatures
    comfort_temp: float
    eco_temp: float
    ice_temp: float

    # User mode
    um_max_temp: float
    um_min_temp: float
    user_mode: bool
    ice_mode: bool

    # Schedule
    schedule: list[str]
    schedule_day: int
    schedule_hour: int

    energy_data: EnergyConsumptionData

    last_sync_datetime_app: datetime
    last_sync_datetime_device: datetime

    hass_available: bool

    def __init__(
        self,
        device_id: str,
        device_info: dict,
        energy_data: EnergyConsumptionData,
        latest_fw: str | None,
    ) -> None:
        """Initialize the device from the rointe's json blob."""
        self.id = device_id
        self.type = device_info["data"]["type"]
        self.product_version = str.lower(device_info["data"]["product_version"])
        self.serialnumber = device_info["serialnumber"]
        self.update_data(device_info, energy_data, latest_fw)

    def update_data(
        self,
        device_info: dict,
        energy_data: EnergyConsumptionData,
        latest_fw: str | None,
    ) -> None:
        """Update the device data from a Json object."""

        data = device_info["data"]
        firmware_data = device_info.get("firmware")

        self.name = data["name"]
        self.nominal_power = int(data["nominal_power"])
        self.power = bool(data["power"])
        self.preset = data["status"]
        self.mode = data["mode"]

        self.temp = float(data["temp"])
        self.temp_calc = float(data["temp_calc"])
        self.temp_probe = float(data["temp_probe"])

        self.comfort_temp = float(data["comfort"])
        self.eco_temp = float(data["eco"])
        self.ice_temp = float(data["ice"])

        # User mode settings are only valid for V2 radiators.
        if self.product_version == "v2":
            self.um_max_temp = float(data["um_max_temp"])
            self.um_min_temp = float(data["um_min_temp"])
            self.user_mode = bool(data["user_mode"])
        else:
            self.user_mode = False

        self.ice_mode = bool(data["ice_mode"])
        self.schedule = data["schedule"]
        self.schedule_day = data["schedule_day"]
        self.schedule_hour = data["schedule_hour"]

        self.energy_data = energy_data

        self.last_sync_datetime_app = datetime.fromtimestamp(
            int(data["last_sync_datetime_app"]) / 1000.0
        )

        self.last_sync_datetime_device = datetime.fromtimestamp(
            int(data["last_sync_datetime_device"]) / 1000.0
        )

        if firmware_data:
            self.firmware_version = firmware_data.get("firmware_version_device")
        else:
            self.firmware_version = None

        self.latest_firmware_version = latest_fw
        self.hass_available = True

    def get_current_schedule_mode(self) -> ScheduleMode:
        """Return the current schedule mode for the device.

        Returns C for Comfort, E for Eco, O for no-schedule
        """
        day_time = utils.now()
        day_of_week = day_time.weekday()  # 0 is Monday
        hour_index = day_time.hour

        current_mode = self.schedule[day_of_week][hour_index]

        if current_mode == "C":
            return ScheduleMode.COMFORT
        elif current_mode == "E":
            return ScheduleMode.ECO
        else:
            return ScheduleMode.NONE

    def user_mode_supported(self) -> bool:
        """Return True if this device supports user mode."""
        return self.product_version == "v2"

    @property
    def rointe_product(self) -> RointeProduct | None:
        """Return the product name."""
        return get_product_by_type_version(self.type, self.product_version)



"""Lib setup"""

from setuptools import find_packages, setup


def requirements():
    """Load requirements"""
    with open("requirements.txt") as fileobj:
        return [line.strip() for line in fileobj]


with open("README.md", encoding="utf-8") as fh:
    doc_long_description = fh.read()

# This call to setup() does all the work
setup(
    name="rointe-sdk",
    version="1.4.0",
    author="tggm",
    description="Python SDK for rointeconnect.com",
    long_description=doc_long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tggm/rointe-sdk",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements(),
)



"""Rointe API Client"""
from __future__ import annotations

import requests
from requests.exceptions import RequestException

from typing import Any, Dict, Optional, List
from collections import namedtuple
from datetime import datetime, timedelta

from .utils import build_update_map
from .device import RointeDevice, ScheduleMode
from .dto import EnergyConsumptionData

from .settings import (
    AUTH_ACCT_INFO_URL,
    AUTH_HOST,
    AUTH_REFRESH_ENDPOINT,
    AUTH_TIMEOUT_SECONDS,
    AUTH_VERIFY_URL,
    ENERGY_STATS_MAX_TRIES,
    FIREBASE_APP_KEY,
    FIREBASE_DEFAULT_URL,
    FIREBASE_DEVICE_DATA_PATH_BY_ID,
    FIREBASE_DEVICE_ENERGY_PATH_BY_ID,
    FIREBASE_DEVICES_PATH_BY_ID,
    FIREBASE_GLOBAL_SETTINGS_PATH,
    FIREBASE_INSTALLATIONS_PATH,
)

ApiResponse = namedtuple("ApiResponse", ["success", "data", "error_message"])


class RointeAPI:
    """Rointe API Communication. Handles low level calls to the API."""

    def __init__(self, username: str, password: str):
        """Initializes the API"""

        self.username = username
        self.password = password

        self.refresh_token = None
        self.auth_token = None
        self.auth_token_expire_date = None
        self.local_id = None

    def initialize_authentication(self) -> ApiResponse:
        """
        Initializes the refresh token and cleans
        the original credentials.
        """

        login_data: ApiResponse = self._login_user()

        if not login_data.success:
            self.auth_token = None
            self.refresh_token = None
            return login_data

        self.auth_token = login_data.data["auth_token"]
        self.refresh_token = login_data.data["refresh_token"]
        self.auth_token_expire_date = login_data.data["expires"]
        self.local_id = login_data.data["local_id"]

        self._clean_credentials()

        return ApiResponse(True, None, None)

    def _clean_credentials(self):
        """Cleans authentication values"""
        self.username = None
        self.password = None

    def is_logged_in(self) -> bool:
        """Check if the login was successful."""
        return self.auth_token is not None and self.refresh_token is not None

    def _ensure_valid_auth(self) -> bool:
        """Ensure there is a valid authentication token present."""

        now = datetime.now()

        if not self.auth_token or (
            self.auth_token_expire_date and self.auth_token_expire_date < now
        ):
            if not self._refresh_token():
                return False

        return True

    def _refresh_token(self) -> bool:
        """Refreshes authentication."""

        payload = {"grant_type": "refresh_token", "refresh_token": self.refresh_token}

        try:
            response = requests.post(
                f"{AUTH_REFRESH_ENDPOINT}?key={FIREBASE_APP_KEY}",
                data=payload,
                timeout=AUTH_TIMEOUT_SECONDS,
            )
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if not response:
            return False

        if response.status_code != 200:
            return False

        response_json = response.json()

        if not response_json or "id_token" not in response_json:
            return False

        self.auth_token = response_json["id_token"]
        self.auth_token_expire_date = datetime.now() + timedelta(
            seconds=int(response_json["expires_in"])
        )
        self.refresh_token = response_json["refresh_token"]

        return True

    def _login_user(self) -> ApiResponse:
        """Log the user in."""

        payload = {
            "email": self.username,
            "password": self.password,
            "returnSecureToken": True,
        }

        try:
            response = requests.post(
                f"{AUTH_HOST}{AUTH_VERIFY_URL}?key={FIREBASE_APP_KEY}",
                data=payload,
                timeout=AUTH_TIMEOUT_SECONDS,
            )
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if response.status_code == 400:
            return ApiResponse(
                False,
                None,
                "Authentication returned 400 (Bad Request)",
            )

        if response.status_code != 200:
            return ApiResponse(
                False, None, "Authentication returned: " + str(response.status_code)
            )

        response_json = response.json()

        if not response_json or "idToken" not in response_json:
            return ApiResponse(
                False, None, "Authentication returned invalid or empty response"
            )

        data = {
            "auth_token": response_json["idToken"],
            "expires": datetime.now()
            + timedelta(seconds=int(response_json["expiresIn"])),
            "refresh_token": response_json["refreshToken"],
            "local_id": response_json["localId"],
        }

        return ApiResponse(True, data, None)

    def get_local_id(self) -> ApiResponse:
        """Retrieve user local_id value."""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        payload = {"idToken": self.auth_token}

        try:
            response = requests.post(
                f"{AUTH_HOST}{AUTH_ACCT_INFO_URL}?key={FIREBASE_APP_KEY}",
                data=payload,
            )
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if not response:
            return ApiResponse(
                False, None, "No response from API call to get_local_id()"
            )

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_local_id() returned {response.status_code}"
            )

        response_json = response.json()

        return ApiResponse(True, response_json["users"][0]["localId"], None)

    def get_installation_devices(self, installation_id: str) -> ApiResponse:
        """Retrieve all devices present in an installation."""
        installation_response = self.get_installation_by_id(installation_id)

        if not installation_response.success:
            return ApiResponse

        detected_devices: List[str] = []

        for zone_key in installation_response.data["zones"]:
            detected_devices.extend(
                self._extract_devices(installation_response.data["zones"].get(zone_key))
            )

        return ApiResponse(True, detected_devices, None)

    def _extract_devices(self, zone_data: dict[str, Any]) -> List[str]:
        """Parses a single zone block recursively."""

        if not zone_data:
            return []

        zone_devices: List[str] = []

        if devices := zone_data.get("devices"):
            zone_devices.extend(list(devices.keys()))

        if sub_zones := zone_data.get("zones"):
            for sub_zone_key in sub_zones:
                zone_devices.extend(self._extract_devices(sub_zones.get(sub_zone_key)))

        return zone_devices

    def get_installation_by_id(self, installation_id: str) -> ApiResponse:
        """Retrieve a specific installation by ID."""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        args = {
            "auth": self.auth_token,
            "orderBy": '"userid"',
            "equalTo": f'"{self.local_id}"',
        }

        url = f"{FIREBASE_DEFAULT_URL}{FIREBASE_INSTALLATIONS_PATH}"

        try:
            response = requests.get(url, params=args)
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if not response:
            return ApiResponse(
                False, "No response from API in get_installation_by_id()"
            )

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_installation_by_id() returned {response.status_code}"
            )

        reponse_json = response.json()

        if len(reponse_json) == 0 or installation_id not in reponse_json:
            return ApiResponse(False, None, "No Rointe installation found.")

        return ApiResponse(True, reponse_json[installation_id], None)

    def get_latest_firmware(self) -> ApiResponse:
        """Retrieves the latest firmware available for each device type"""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        url = f"{FIREBASE_DEFAULT_URL}{FIREBASE_GLOBAL_SETTINGS_PATH}"
        args = {"auth": self.auth_token}

        try:
            response = requests.get(
                url,
                params=args,
            )
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if not response:
            return ApiResponse(False, "No response from API in get_latest_firmware()")

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_latest_firmware() returned {response.status_code}"
            )

        data = response.json()

        if len(data) == 0:
            return ApiResponse(False, None, "Global Settings is empty.")

        return ApiResponse(True, build_update_map(data), None)

    def get_installations(self) -> ApiResponse:
        """Retrieve the client's installations."""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        args = {
            "auth": self.auth_token,
            "orderBy": '"userid"',
            "equalTo": f'"{self.local_id}"',
        }
        url = f"{FIREBASE_DEFAULT_URL}{FIREBASE_INSTALLATIONS_PATH}"

        try:
            response = requests.get(url, params=args)
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if not response:
            return ApiResponse(False, "No response from API in get_installations()")

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_installations() returned {response.status_code}"
            )

        reponse_json = response.json()

        if len(reponse_json) == 0:
            return ApiResponse(False, None, "No Rointe installations found.")

        installations = {}

        for key in reponse_json.keys():
            installations[key] = reponse_json[key]["location"]

        return ApiResponse(True, installations, None)

    def get_device(self, device_id: str) -> ApiResponse:
        """Retrieve device data."""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        args = {"auth": self.auth_token}

        try:
            response = requests.get(
                "{}{}".format(
                    FIREBASE_DEFAULT_URL, FIREBASE_DEVICES_PATH_BY_ID.format(device_id)
                ),
                params=args,
            )
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if not response:
            return ApiResponse(False, "No response from API in get_device()")

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_device() returned {response.status_code}"
            )

        return ApiResponse(True, response.json(), None)

    def get_latest_energy_stats(self, device_id: str) -> ApiResponse:
        """Retrieve the latest energy consumption values."""

        result: EnergyConsumptionData
        now = datetime.now()

        # Attempt to retrieve the latest value. If not found, go back one hour. Max 5 tries.
        attempts = ENERGY_STATS_MAX_TRIES
        target_date = now.replace(
            minute=0, second=0, microsecond=0
        )  # Strip minutes, seconds and microseconds.

        while attempts > 0:
            result: ApiResponse = self._retrieve_hour_energy_stats(
                device_id, target_date
            )

            if result.error_message == "No energy stats found.":
                # Try again.
                attempts = attempts - 1
                target_date = target_date - timedelta(hours=1)
            else:
                # It's either a success or an error message, return the ApiResponse.
                return result

        return ApiResponse(False, None, "Max tries exceeded.")

    def _retrieve_hour_energy_stats(
        self, device_id: str, target_date: datetime
    ) -> ApiResponse:
        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        # Sample URL /history_statistics/device_id/daily/2022/01/21/energy/010000.json
        args = {"auth": self.auth_token}
        url = "{}{}{}/energy/{}0000.json".format(
            FIREBASE_DEFAULT_URL,
            FIREBASE_DEVICE_ENERGY_PATH_BY_ID.format(device_id),
            target_date.strftime("%Y/%m/%d"),
            target_date.strftime("%H"),
        )

        try:
            response = requests.get(url, params=args)
        except RequestException as e:
            return ApiResponse(False, None, f"Network error {e}")

        if not response:
            return ApiResponse(
                False, "No response from API in _retrieve_hour_energy_stats()"
            )

        if response.status_code != 200:
            return ApiResponse(
                False,
                None,
                f"_retrieve_hour_energy_stats() returned {response.status_code}",
            )

        response_json = response.json()

        if not response_json or len(response_json) == 0:
            return ApiResponse(False, None, "No energy stats found.")

        data = EnergyConsumptionData(
            created=datetime.now,
            start=target_date,
            end=target_date + timedelta(hours=1),
            kwh=float(response_json["kw_h"]),
            effective_power=float(response_json["effective_power"]),
        )

        return ApiResponse(True, data, None)

    def set_device_temp(self, device: RointeDevice, new_temp: float) -> ApiResponse:
        """Set the device target temperature."""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        device_id = device.id
        args = {"auth": self.auth_token}
        body = {"temp": new_temp, "mode": "manual", "power": True}

        url = "{}{}".format(
            FIREBASE_DEFAULT_URL, FIREBASE_DEVICE_DATA_PATH_BY_ID.format(device_id)
        )

        return self._send_patch_request(url, args, body)

    def set_device_preset(self, device: RointeDevice, preset_mode: str) -> ApiResponse:
        """Set the preset."""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        device_id = device.id
        args = {"auth": self.auth_token}
        body: Dict[str, Any] = {}

        url = "{}{}".format(
            FIREBASE_DEFAULT_URL, FIREBASE_DEVICE_DATA_PATH_BY_ID.format(device_id)
        )

        if preset_mode == "comfort":
            body = {
                "power": True,
                "mode": "manual",
                "temp": device.comfort_temp,
                "status": "comfort",
            }

        elif preset_mode == "eco":
            body = {
                "power": True,
                "mode": "manual",
                "temp": device.eco_temp,
                "status": "eco",
            }
        elif preset_mode == "ice":
            body = {
                "power": True,
                "mode": "manual",
                "temp": device.ice_temp,
                "status": "ice",
            }

        return self._send_patch_request(url, args, body)

    def set_device_mode(self, device: RointeDevice, hvac_mode: str) -> ApiResponse:
        """Set the HVAC mode."""

        if not self._ensure_valid_auth():
            return ApiResponse(False, None, "Invalid authentication.")

        device_id = device.id
        args = {"auth": self.auth_token}
        body: Dict[str, Any] = {}

        url = "{}{}".format(
            FIREBASE_DEFAULT_URL, FIREBASE_DEVICE_DATA_PATH_BY_ID.format(device_id)
        )

        if hvac_mode == "off":
            # This depends if the device is in Auto or Manual modes.
            if device.mode == "auto":
                body = {"power": False, "mode": "auto", "status": "off"}
                return self._send_patch_request(url, args, body)
            else:
                # When turning the device off, we need to set the temperature first.
                set_mode_response = self._send_patch_request(url, args, {"temp": 20})

                if not set_mode_response.success:
                    return set_mode_response

                # Then we can turn the device off.
                body = {"power": False, "mode": "manual", "status": "off"}
                return self._send_patch_request(url, args, body)

        elif hvac_mode == "heat":
            set_mode_response = self._send_patch_request(
                url, args, {"temp": device.comfort_temp}
            )

            if not set_mode_response.success:
                return set_mode_response

            body = {"mode": "manual", "power": True, "status": "none"}
            return self._send_patch_request(url, args, body)

        elif hvac_mode == "auto":
            current_mode: ScheduleMode = device.get_current_schedule_mode()

            # When changing modes we need to send the proper
            # temperature also.
            if current_mode == ScheduleMode.COMFORT:
                body = {"temp": device.comfort_temp}
            elif current_mode == ScheduleMode.ECO:
                body = {"temp": device.eco_temp}
            elif device.ice_mode:
                body = {"temp": device.ice_temp}
            else:
                body = {"temp": 20}

            set_mode_response = self._send_patch_request(url, args, body)

            if not set_mode_response.success:
                return set_mode_response

            # and then set AUTO mode.
            request_mode_status = self._send_patch_request(
                url, args, {"mode": "auto", "power": True}
            )

            return request_mode_status
        else:
            return ApiResponse(False, None, f"Invalid HVAC Mode {hvac_mode}.")

    def _send_patch_request(
        self,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        body=None,
    ) -> ApiResponse:
        """Send a patch request."""

        body["last_sync_datetime_app"] = round(datetime.now().timestamp() * 1000)

        try:
            response = requests.patch(
                url,
                params=params,
                json=body,
            )
        except RequestException as e:
            return ApiResponse(False, None, f"Communications error {e}")

        if not response:
            return ApiResponse(False, None, None)

        if response.status_code != 200:
            return ApiResponse(False, None, None)

        return ApiResponse(True, None, None)
